# This is a workaround to create an alias for the nxlib package with the old
# package name ensenso_nxlib.
# https://stackoverflow.com/questions/56559817/how-do-i-alias-a-python-module-at-packaging-time

import sys

import nxlib
from nxlib import *


sys.modules["ensenso_nxlib"] = nxlib



import atexit
import os
from ctypes import CDLL, POINTER
from ctypes import byref, c_char_p, c_double, c_int32, c_void_p, cast

import _ctypes

from nxlib import _helper as helper
from nxlib.constants import *
from nxlib.exception import NxLibError, NxLibException


__all__ = [
    "close_tcp_port",
    "connect",
    "disconnect",
    "erase",
    "finalize",
    "get_binary",
    "get_binary_info",
    "get_bool",
    "get_count",
    "get_debug_buffer",
    "get_debug_messages",
    "get_double",
    "get_int",
    "get_json",
    "get_json_meta",
    "get_name",
    "get_string",
    "get_type",
    "initialize",
    "is_current_lib_remote",
    "load_lib",
    "load_remote_lib",
    "make_unique_item",
    "open_tcp_port",
    "set_binary",
    "set_binary_formatted",
    "set_bool",
    "set_double",
    "set_int",
    "set_json",
    "set_null",
    "set_string",
    "translate_error_code",
    "wait_for_bool_value",
    "wait_for_change",
    "wait_for_double_value",
    "wait_for_int_value",
    "wait_for_string_value",
    "wait_for_type",
]


class _Nxlib():
    def __init__(self):
        self.is_remote = False
        self.lib_object = None
        atexit.register(self.reset)

    def reset(self):
        if self.lib_object is None:
            return
        try:
            if self.is_remote:
                disconnect()
            else:
                finalize()
        except NxLibException:
            # Ignore any errors while finalizing the NxLib. We will unload the
            # library anyway.
            pass

        # ctypes does not unload the library even when the CDLL object gets
        # garbage collected, so we unload it manually.
        if os.name == "posix":
            _ctypes.dlclose(self.lib_object._handle)
        elif os.name == "nt":
            _ctypes.FreeLibrary(self.lib_object._handle)

        del self.lib_object
        self.lib_object = None

    def __del__(self):
        try:
            self.reset()
        except NxLibException:
            # Sometimes we cannot clean up the library when unloading the
            # module, because some things that are needed for the destruction
            # are already unloaded. Ignore any errors this might cause.
            pass


# testing flag
__nx_testing__ = False

# global placeholder for the nxlib_instance
_nxlib = None

# Instance ID that is incremented with each `initialize`. Used to determine
# which instance a temporary node belongs to.
_nxlib_id = 0


def _get_lib(path=None):
    # will default load the normal nxlib, if there has been no
    # load_remote_lib beforehand!
    global _nxlib

    if _nxlib is None:
        _nxlib = _Nxlib()

    if _nxlib.lib_object is None:
        if _nxlib.is_remote:
            load_remote_lib(path)
        else:
            load_lib(path)
    return _nxlib.lib_object


def is_current_lib_remote():
    """
    Checks whether the currently loaded NxLib library is a remote NxLib.

    Returns:
        bool: True if the currently loaded NxLib library is a remote NxLib.
        False otherwise.
    """
    # For Testing and debug purpose
    global _nxlib
    if _nxlib is None:
        return False
    return _nxlib.is_remote


def load_lib(path=None):
    """
    Loads the default or given NxLib.

    Args:
        path (str, optional): Filepath to the shared library. Defaults to None.

    Raises:
        `nxlib.NxLibError`: If the shared library could not be loaded.
    """
    global _nxlib

    if _nxlib is None:
        _nxlib = _Nxlib()

    _nxlib.reset()

    if path is None:
        _nxlib.lib_object = CDLL(helper.get_lib_path())
    else:
        _nxlib.lib_object = CDLL(path)
    _nxlib.is_remote = False
    if _nxlib.lib_object is None:
        raise NxLibError("Could not load shared library")


def load_remote_lib(path=None):
    """
    Loads the default or given remote NxLib.

    Args:
        path (str, optional): Filepath to the shared remote library.
            Defaults to None.

    Raises:
        `nxlib.NxLibError`: If the shared remote library could not be loaded.
    """
    global _nxlib

    if _nxlib is None:
        _nxlib = _Nxlib()

    _nxlib.reset()

    if path is None:
        _nxlib.lib_object = CDLL(helper.get_lib_path(is_remote_lib=True))
    else:
        _nxlib.lib_object = CDLL(path)
    _nxlib.is_remote = True
    if _nxlib.lib_object is None:
        raise NxLibError("Could not load shared remote library")


def _check_return_code(path="", return_code=c_int32(0)):
    if isinstance(return_code, c_int32):
        return_code = return_code.value
    if return_code != NXLIB_OPERATION_SUCCEEDED:
        raise NxLibException(path, return_code)


def _check_string(s):
    if s is not None:
        return s.decode()
    return None


def _set(f, path, value):
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    f(byref(return_code), path, value)
    _check_return_code(path, return_code)


def set_null(path):
    f = _get_lib().nxLibSetNull
    f.argtypes = [POINTER(c_int32), c_char_p]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    f(byref(return_code), path)
    _check_return_code(path, return_code)


def set_int(path, value):
    f = _get_lib().nxLibSetInt
    f.argtypes = [POINTER(c_int32), c_char_p, c_int32]
    _set(f, path, value)


def set_double(path, value):
    f = _get_lib().nxLibSetDouble
    f.argtypes = [POINTER(c_int32), c_char_p, c_double]
    _set(f, path, value)


def set_bool(path, value):
    f = _get_lib().nxLibSetBool
    f.argtypes = [POINTER(c_int32), c_char_p, c_int32]
    _set(f, path, value)


def set_string(path, value):
    f = _get_lib().nxLibSetString
    f.argtypes = [POINTER(c_int32), c_char_p, c_char_p]
    value = helper.fix_string_encoding(value)
    _set(f, path, value)


def set_json(path, value, only_writeable_nodes=False):
    f = _get_lib().nxLibSetJson
    f.argtypes = [POINTER(c_int32), c_char_p, c_char_p, c_int32]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    value = helper.fix_string_encoding(value)
    f(byref(return_code), path, value, only_writeable_nodes)
    _check_return_code(path, return_code)


def set_binary(path, buffer, buffer_size):
    f = _get_lib().nxLibSetBinary
    f.argtypes = [POINTER(c_int32), c_char_p, POINTER(c_void_p), c_int32]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    buffer = cast(buffer, POINTER(c_void_p))
    f(byref(return_code), path, buffer, buffer_size)
    _check_return_code(path, return_code)


def set_binary_formatted(path, buffer, width, height, channel_count,
                         bytes_per_element, is_float):
    f = _get_lib().nxLibSetBinaryFormatted
    f.argtypes = [
        POINTER(c_int32), c_char_p,
        POINTER(c_void_p), c_int32, c_int32, c_int32, c_int32, c_int32
    ]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    buffer = cast(buffer, POINTER(c_void_p))
    f(byref(return_code), path, buffer, width, height, channel_count,
      bytes_per_element, is_float)
    _check_return_code(path, return_code)


def get_binary(path, buffer, buffer_size):
    f = _get_lib().nxLibGetBinary
    f.argtypes = [
        POINTER(c_int32), c_char_p,
        POINTER(c_void_p), c_int32,
        POINTER(c_int32),
        POINTER(c_double)
    ]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    buffer = cast(buffer, POINTER(c_void_p))
    bytes_copied = c_int32(0)
    timestamp = c_double(0)
    f(byref(return_code), path, buffer, buffer_size, byref(bytes_copied),
      byref(timestamp))
    _check_return_code(path, return_code)
    return bytes_copied.value, timestamp.value


def get_binary_info(path):
    f = _get_lib().nxLibGetBinaryInfo
    f.argtypes = [
        POINTER(c_int32), c_char_p,
        POINTER(c_int32),
        POINTER(c_int32),
        POINTER(c_int32),
        POINTER(c_int32),
        POINTER(c_int32),
        POINTER(c_double)
    ]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    width = c_int32(0)
    height = c_int32(0)
    channel_count = c_int32(0)
    bytes_per_element = c_int32(0)
    is_float = c_int32(0)
    timestamp = c_double(0)
    f(byref(return_code), path, byref(width), byref(height),
      byref(channel_count), byref(bytes_per_element), byref(is_float),
      byref(timestamp))
    _check_return_code(path, return_code)
    return (width.value, height.value, channel_count.value,
            bytes_per_element.value, is_float.value == 1, timestamp.value)


def has_path(path):
    return_code = c_int32(0)
    type_ = get_type(path, return_code)
    if return_code.value == NXLIB_OPERATION_SUCCEEDED:
        # Item type can still be invalid.
        return type_ != NXLIB_ITEM_TYPE_INVALID
    elif return_code.value == NXLIB_ITEM_INEXISTENT:
        return False
    # Otherwise raise an execption.
    _check_return_code(path, return_code)


def _get(f, path, return_code=None):
    return_code_ = c_int32(0)
    path = helper.fix_string_encoding(path)
    result = f(byref(return_code_), path)
    if return_code is not None:
        return_code = return_code_
    else:
        _check_return_code(path, return_code_)
    return result


def get_type(path, return_code=None):
    f = _get_lib().nxLibGetType
    f.restype = c_int32
    f.argtypes = [POINTER(c_int32), c_char_p]
    return _get(f, path, return_code)


def get_int(path):
    f = _get_lib().nxLibGetInt
    f.restype = c_int32
    f.argtypes = [POINTER(c_int32), c_char_p]
    return _get(f, path)


def get_bool(path):
    f = _get_lib().nxLibGetBool
    f.restype = c_int32
    f.argtypes = [POINTER(c_int32), c_char_p]
    return bool(_get(f, path))


def get_count(path):
    f = _get_lib().nxLibGetCount
    f.restype = c_int32
    f.argtypes = [POINTER(c_int32), c_char_p]
    return _get(f, path)


def get_double(path):
    f = _get_lib().nxLibGetDouble
    f.restype = c_double
    f.argtypes = [POINTER(c_int32), c_char_p]
    return _get(f, path)


def get_string(path):
    f = _get_lib().nxLibGetString
    f.restype = c_char_p
    f.argtypes = [POINTER(c_int32), c_char_p]
    return _check_string(_get(f, path))


def get_name(path):
    f = _get_lib().nxLibGetName
    f.restype = c_char_p
    f.argtypes = [POINTER(c_int32), c_char_p]
    return _check_string(_get(f, path))


def get_json(path, pretty_print, number_precision, scientific_number_format):
    f = _get_lib().nxLibGetJson
    f.restype = c_char_p
    f.argtypes = [POINTER(c_int32), c_char_p, c_int32, c_int32, c_int32]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    result = f(byref(return_code), path, pretty_print, number_precision,
               scientific_number_format)
    _check_return_code(path, return_code)
    return _check_string(result)


def get_json_meta(path, num_levels, pretty_print, number_precision,
                  scientific_number_format):
    f = _get_lib().nxLibGetJsonMeta
    f.restype = c_char_p
    f.argtypes = [
        POINTER(c_int32), c_char_p, c_int32, c_int32, c_int32, c_int32
    ]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    result = f(byref(return_code), path, num_levels, pretty_print,
               number_precision, scientific_number_format)
    _check_return_code(path, return_code)
    return _check_string(result)


def erase(path):
    f = _get_lib().nxLibErase
    f.argtypes = [POINTER(c_int32), c_char_p]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    f(byref(return_code), path)
    if return_code.value == NXLIB_ITEM_INEXISTENT:
        return
    _check_return_code(path, return_code)


def wait_for_change(path):
    f = _get_lib().nxLibWaitForChange
    f.argtypes = [POINTER(c_int32), c_char_p]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    f(byref(return_code), path)
    _check_return_code(path, return_code)


def wait_for_type(path, awaited_type, wait_for_equal):
    f = _get_lib().nxLibWaitForType
    f.argtypes = [POINTER(c_int32), c_char_p, c_int32, c_int32]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    f(byref(return_code), path, awaited_type, wait_for_equal)
    _check_return_code(path, return_code)


def wait_for_int_value(path, value, wait_for_equal):
    f = _get_lib().nxLibWaitForIntValue
    f.argtypes = [POINTER(c_int32), c_char_p, c_int32, c_int32]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    f(byref(return_code), path, value, wait_for_equal)
    _check_return_code(path, return_code)


def wait_for_string_value(path, value, wait_for_equal):
    f = _get_lib().nxLibWaitForStringValue
    f.argtypes = [POINTER(c_int32), c_char_p, c_char_p, c_int32]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    value = helper.fix_string_encoding(value)
    f(byref(return_code), path, value, wait_for_equal)
    _check_return_code(path, return_code)


def wait_for_double_value(path, value, wait_for_equal):
    f = _get_lib().nxLibWaitForDoubleValue
    f.argtypes = [POINTER(c_int32), c_char_p, c_double, c_int32]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    f(byref(return_code), path, value, wait_for_equal)
    _check_return_code(path, return_code)


def wait_for_bool_value(path, value, wait_for_equal):
    f = _get_lib().nxLibWaitForBoolValue
    f.argtypes = [POINTER(c_int32), c_char_p, c_int32, c_int32]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    f(byref(return_code), path, value, wait_for_equal)
    _check_return_code(path, return_code)


def make_unique_item(path, item_name):
    f = _get_lib().nxLibMakeUniqueItem
    f.restype = c_char_p
    f.argtypes = [POINTER(c_int32), c_char_p, c_char_p]
    return_code = c_int32(0)
    path = helper.fix_string_encoding(path)
    item_name = helper.fix_string_encoding(item_name)
    result = f(byref(return_code), path, item_name)
    _check_return_code(return_code=return_code)
    return _check_string(result)


def translate_error_code(error_code):
    """
    Returns the corresponding error text for the given NxLib error code.

    Args:
        error_code (int): The error code.

    Returns:
        str: The corresponding error text.
    """
    if error_code < 0 or error_code > len(NX_ERRORS):
        return ''
    return NX_ERRORS[error_code]


def get_debug_messages():
    f = _get_lib().nxLibGetDebugMessages
    f.restype = c_char_p
    f.argtypes = [POINTER(c_int32)]
    return_code = c_int32(0)
    result = f(byref(return_code))
    _check_return_code(return_code=return_code)
    return _check_string(result)


def get_debug_buffer(destination_buffer, buffer_size, clear_read):
    """
    Copy n=buffer_size bytes of the NxLib debug buffer into the given
    destination buffer and clear the debug buffer if desired.

    Args:
        destination_buffer (``Object``): The byte buffer the debug buffer's
            content gets written to.
        buffer_size (int): The buffer size.
        clear_read (bool): If True, the buffer will be cleared after reading.

    Raises:
        NxLibException: If the wrapped NxLib function returned with an error.

    Returns:
        tuple containing

        * bytes_read (int): The number of bytes read from the debug buffer.
        * bytes_remaining (int): The number of bytes remaining in the debug
          buffer.
        * overflow (bool): True if the debug buffer overflowed before it was
          read. False otherwise.
    """
    f = _get_lib().nxLibGetDebugBuffer
    f.restype = c_int32
    f.argtypes = [
        POINTER(c_int32),
        POINTER(c_void_p), c_int32,
        POINTER(c_int32), c_int32
    ]
    return_code = c_int32(0)
    destination_buffer = cast(destination_buffer, POINTER(c_void_p))
    bytes_remaining = c_int32(0)
    clear_read = c_int32(clear_read)
    bytes_read = f(byref(return_code), destination_buffer, buffer_size,
                   byref(bytes_remaining), clear_read)
    overflow = False
    if return_code.value == NXLIB_DEBUG_MESSAGE_OVERFLOW:
        overflow = True
    else:
        _check_return_code(return_code=return_code)
    return bytes_read, bytes_remaining.value, overflow


def _log_level_str_to_int(level):
    if level not in NX_LOG_LEVELS:
        return 0
    return NX_LOG_LEVELS[level]


def open_debug_block(block_name, level):
    f = _get_lib().nxLibOpenDebugBlock
    f.argtypes = [POINTER(c_int32), c_char_p, c_int32]
    return_code = c_int32(0)
    block_name = helper.fix_string_encoding(block_name)
    level = _log_level_str_to_int(level)
    f(byref(return_code), block_name, level)
    _check_return_code(return_code=return_code)


def close_debug_block():
    f = _get_lib().nxLibCloseDebugBlock
    f.argtypes = [POINTER(c_int32)]
    return_code = c_int32(0)
    f(byref(return_code))
    _check_return_code(return_code=return_code)


def write_debug_message(message):
    f = _get_lib().nxLibWriteDebugMessage
    f.argtypes = [POINTER(c_int32), c_char_p]
    return_code = c_int32(0)
    message = helper.fix_string_encoding(message)
    f(byref(return_code), message)
    _check_return_code(return_code=return_code)


def initialize(wait_for_initial_camera_refresh=True, path=None):
    """
    Explicitly initializes the library and starts to enumerate the connected
    cameras. When omitting this function call, the library is initialized on
    first access to the tree.

    Note:
        After :func:`finalize` was called, the library will no longer be
        initialized automatically when you access the tree. You can reinitialize
        it by explicitly calling :func:`~initialize`.

    Args:
        wait_for_initial_camera_refresh (bool, optional): Specifies whether to
            wait for all cameras to be enumerated or to return immediately.
            Defaults to True.

    Raises:
        `nxlib.NxLibError`: If the currently loaded NxLib library is a remote
            NxLib.
    """
    if is_current_lib_remote():
        raise NxLibError("Library is a remote NxLib. Only normal NxLib "
                         "instances can use initialize.")
    f = _get_lib(path).nxLibInitialize
    f.argtypes = [POINTER(c_int32), c_int32]
    return_code = c_int32(0)
    f(byref(return_code), wait_for_initial_camera_refresh)
    _check_return_code(return_code=return_code)
    global _nxlib_id
    _nxlib_id += 1


def finalize():
    """
    Explicitly closes the library, terminating all internal threads and freeing
    allocated memory. It is important to explicitly call :func:`~finalize`
    before unloading the NxLib library when your process is not terminating
    afterwards, because Windows doesn't allow to cleanly exit threads during DLL
    unload.

    Raises:
        `nxlib.NxLibError`: If the currently loaded NxLib library is a remote
            NxLib.
    """
    if _nxlib is None or _nxlib.lib_object is None:
        return
    if is_current_lib_remote():
        raise NxLibError("Library is a remote NxLib. Only normal NxLib "
                         "instances can use finalize.")
    f = _get_lib().nxLibFinalize
    f.argtypes = [POINTER(c_int32)]
    return_code = c_int32(0)
    f(byref(return_code))
    _check_return_code(return_code=return_code)


def open_tcp_port(port_number=0, opened_port=0):
    """
    Opens a TCP port on which a remote NxLib can connect to the current NxLib
    instance.

    Args:
        port_number (int, optional): The port to be opened. Specify 0 here to
            automatically choose a port in the range 24000 to 25000.
            Defaults to 0.
        opened_port (int, optional): The variable receiving the opened port
            number. Defaults to 0.

    Raises:
        `nxlib.NxLibError`: If the currently loaded NxLib library is a remote
            NxLib.
    """
    if is_current_lib_remote():
        raise NxLibError("Library is a remote NxLib. Only normal NxLib "
                         "instances are allowed to open TCP ports.")
    f = _get_lib().nxLibOpenTcpPort
    f.argtypes = [POINTER(c_int32), c_int32, c_int32]
    return_code = c_int32(0)
    f(byref(return_code), port_number, opened_port)
    _check_return_code(return_code=return_code)
    if __nx_testing__:
        print(port_number)


def close_tcp_port():
    """
    Disconnects all connected NxLibRemote instances and closes the opened TCP
    port, if any.

    Raises:
        `nxlib.NxLibError`: If the currently loaded NxLib library is a remote
            NxLib.
    """
    if is_current_lib_remote():
        raise NxLibError("Library is a remote NxLib. Only normal NxLib "
                         "instances are allowed to close TCP ports.")
    f = _get_lib().nxLibCloseTcpPort
    return_code = c_int32(0)
    f.argtypes = [POINTER(c_int32)]
    f(byref(return_code))
    _check_return_code(return_code=return_code)


def connect(hostname, port):
    """
    Remote NxLib specific function that opens a connection to the remote NxLib
    with the given hostname and port.

    Args:
        hostname (str): The hostname of the remote NxLib to connect to.
        port (int): The port of the remote NxLib to connect to.

    Raises:
        `nxlib.NxLibError`: If the currently loaded NxLib library is not a
            remote NxLib.
    """
    if not is_current_lib_remote():
        raise NxLibError("Cannot use connect function from a normal NxLib.")
    f = _get_lib().nxLibConnect
    f.argtypes = [POINTER(c_int32), c_char_p, c_int32]
    return_code = c_int32(0)
    hostname = helper.fix_string_encoding(hostname)
    f(byref(return_code), hostname, port)
    _check_return_code(return_code=return_code)


def disconnect():
    """
    Remote NxLib specific function that closes the connection to the remote
    NxLib.

    Raises:
        `nxlib.NxLibError`: If the currently loaded NxLib library is not a
            remote NxLib.
    """
    if not is_current_lib_remote():
        raise NxLibError("Cannot use disconnenct function from a normal "
                         "NxLib.")
    f = _get_lib().nxLibDisconnect
    f.argtypes = [POINTER(c_int32)]
    return_code = c_int32(0)
    f(byref(return_code))
    _check_return_code(return_code=return_code)



import nxlib.api as api
from nxlib.constants import *
from nxlib.exception import NxLibException, _handle_exception
from nxlib.item import NxLibItem

__all__ = ["NxLibCommand"]


class NxLibCommand:
    """
    This class simplifies the execution of NxLib commands. The
    :meth:`~parameters` and :meth:`~result` methods provide access to the
    corresponding ``Parameters`` and ``Result`` nodes of the tree. The
    :meth:`~execute` method starts the command by writing the command name to
    the tree and allows for synchronous and asynchronous command execution.

    Args:
        command_name (str): Name of the command to execute.
        node_name (str, optional): Name of the execution ``slot`` the command
            should run in. By default the class will create a temporary slot
            used only by this instance of :class:`NxLibCommand`, which is
            automatically deleted in the destructor.
    """
    def __init__(self, command_name, node_name=None, params={}):
        self._command_name = command_name
        self._remove_slot_on_destruction = False
        self._nxlib_id = api._nxlib_id

        if node_name is None:
            self._create_temporary_slot()
        else:
            self._command_item = NxLibItem()[ITM_EXECUTE][node_name]

        self._set_parameters(self.parameters(), params)

    def _set_parameters(self, node, params):
        for key, value in params.items():
            if type(value) != dict:
                node[key] = value
            else:
                self._set_parameters(node[key], value)

    def _belongs_to_current_lib(self):
        return self._nxlib_id == api._nxlib_id

    def _remove_slot(self):
        if self._remove_slot_on_destruction and self._belongs_to_current_lib():
            try:
                self._command_item.erase()
            except NxLibException as e:
                if e.get_error_code() == NXLIB_INITIALIZATION_NOT_ALLOWED:
                    pass
                else:
                    raise e
        self._remove_slot_on_destruction = False

    def __del__(self):
        self._remove_slot()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._remove_slot()
        return _handle_exception(exc_type, exc_value, exc_tb)

    def _create_temporary_slot(self, base_name=None):
        self._command_item = NxLibItem()[ITM_EXECUTE].make_unique_item(
            base_name)
        self._remove_slot_on_destruction = True

    def slot(self):
        """
        The slot, i.e. the NxLibItem, the NxLibCommand is running on.

        Returns:
            ``NxLibItem``: The NxLibItem the command runs on.
        """
        return self._command_item

    def parameters(self):
        """
        The command's ``Parameters`` node, that will be fetched before the
        command executes.

        Returns:
            ``NxLibItem``: The NxLibItem in which the parameters are stored.
        """
        return self.slot()[ITM_PARAMETERS]

    def result(self):
        """
        The command's ``Result`` node after the command finishes.

        Returns:
            ``NxLibItem``: The NxLibItem in which the results are stored.
        """
        return self.slot()[ITM_RESULT]

    def successful(self):
        """
        Checks whether the previous command execution was successful.

        Returns:
            bool: True if the previous NxLib command execution was successful,
            i.e. there is no ``ErrorSymbol`` node under the ``Result`` node.
            False otherwise.
        """
        return not self.result()[ITM_ERROR_SYMBOL].exists()

    def execute(self, command_name=None, wait=True):
        """
        Executes the current command.

        Args:
            command_name (str, optional): Name of the command to execute. This
                overwrites the command name from the constructor. Defaults to
                None.
            wait (bool, optional): If True, the function waits until execution
                of the command is finished and throws an NxLibException if the
                command finished with an error. Defaults to True.
        """
        if not command_name:
            command_name = self._command_name

        function_item = self.slot()[ITM_COMMAND]
        function_item.set_string(command_name)

        if wait:
            function_item.wait_for_type(NXLIB_ITEM_TYPE_NULL, True)
            self.assert_successful()

    def finished(self):
        """
        Checks whether the command execution has already finished.

        Returns:
            bool: True, if NxLib is not currently executing a command, i.e. the
            ``Command`` node is not exisintg or is Null. False otherwise.
        """
        cmd_node = self.slot()[ITM_COMMAND]
        return (not cmd_node.exists() or cmd_node.is_null())

    def assert_successful(self):
        """
        Checks whether the previous command execution is finished and was
        successful.

        Raises:
            NxLibException: If the command execution was not successful.
        """
        if not self.finished() or not self.successful():
            raise NxLibException(self.slot().path, NXLIB_EXECUTION_FAILED, self)



# In order so one can just call from nxlib import NxLibException (for example)
# flake8: noqa
from .command import NxLibCommand
from .context import (
	NxLib,
	NxLibRemote,
	Camera,
	MonoCamera,
	StereoCamera,
	StructuredLightCamera)
from .exception import NxLibError, NxLibException
from .item import NxLibItem
from .log import NxDebugBlock, NxLog



from nxlib import api
from nxlib.command import NxLibCommand
from nxlib.constants import *
from nxlib.exception import _handle_exception
from nxlib.item import NxLibItem


class NxLib:
    """
    This class offers a simple to use interface for interacting with a normal
    NxLib. It implements the context manager protocol and thus can be used in a
    ``with`` statement, which automatically initializes the NxLib and takes care
    of the exception handling.

    Args:
        path (str, optional): The path to the NxLib shared library.
            Defaults to None.
    """
    def __enter__(self, path=None):
        api.initialize(path=path)
        return api

    def __exit__(self, exc_type, exc_value, exc_tb):
        return _handle_exception(exc_type, exc_value, exc_tb)


class NxLibRemote:
    """
    This class offers a simple to use interface for interacting with a remote
    NxLib. It implements the context manager protocol and thus can be used
    in a ``with`` statement, which automatically loads the remote NxLib,
    connects to the given hostname (and port) when entering the scope and
    automatically disconnects when exiting the scope. It also takes care of the
    exception handling.

    Args:
        hostname (str): The hostname of the remote NxLib.
        port (int): The port number of the remote NxLib.

    """
    def __init__(self, hostname, port):
        self._hostname = hostname
        self._port = int(port)

    def __enter__(self):
        api.load_remote_lib()
        api.connect(self._hostname, self._port)
        return api

    def __exit__(self, exc_type, exc_value, exc_tb):
        api.disconnect()
        return _handle_exception(exc_type, exc_value, exc_tb)


class Camera:
    @classmethod
    def from_serial(cls, serial, expected_types=None, open_params={}):
        if expected_types is None:
            expected_types = [VAL_MONOCULAR, VAL_STEREO, VAL_STRUCTURED_LIGHT]
        camera = Camera._get_camera_node(serial)
        camera_type = camera[ITM_TYPE].as_string()
        if camera_type not in expected_types:
            raise CameraTypeError(f"{serial} is of type {camera_type}, "
                f"expected one of {expected_types}")
        if camera_type == VAL_MONOCULAR:
            return MonoCamera(serial, open_params)
        elif camera_type == VAL_STEREO:
            return StereoCamera(serial, open_params)
        elif camera_type == VAL_STRUCTURED_LIGHT:
            return StructuredLightCamera(serial, open_params)

    def __init__(self, serial, open_params):
        self._serial = serial
        self._open_params = open_params
        self._node = Camera._get_camera_node(serial)

    def __getitem__(self, value):
        """
        The ``[]`` access operator.

        Args:
            value (int, str, bool or float): The value to access.

        Returns:
            ``NxLibItem``: The resulting node.
        """
        return self._node[value]

    def get_node(self):
        """
        Get the camera tree node of the stereo camera the context object opens
        and represents.

        Returns:
            `NxLibItem`: The camera node of the stereo camera.
        """
        return self._node

    def capture(self):
        """ Capture the image(s). """
        self._execute(CMD_CAPTURE)

    def rectify(self):
        """
        Rectify the captured images (requires :meth:`capture` to be called
        first). Use this method only if you want to have the rectified raw
        images and no further data.
        """
        self._execute(CMD_RECTIFY_IMAGES)

    def __enter__(self):
        if self._node[ITM_STATUS][ITM_AVAILABLE].as_bool() is False:
            raise CameraOpenError(f"{self._serial} not available")
        self._execute(CMD_OPEN, self._open_params)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._execute(CMD_CLOSE)
        return _handle_exception(exc_type, exc_value, exc_tb)

    def _execute(self, command_name, params={}):
        cmd = NxLibCommand(command_name, params=params)
        cmd.parameters()[ITM_CAMERAS] = self._serial
        cmd.execute()

    def _check_type(self, expected_type):
        camera_type = self._node[ITM_TYPE].as_string()
        if camera_type != expected_type:
            raise CameraTypeError(f"{self._serial} is of type {camera_type}, "
                                  f"expected {expected_type}")

    @classmethod
    def _get_camera_node(cls, serial):
        cameras = NxLibItem()[ITM_CAMERAS][ITM_BY_SERIAL_NO]
        for i in range(cameras.count()):
            if cameras[i].name() == serial:
                return cameras[i]
        raise CameraNotFoundError(f"No camera found for serial {serial}")


class MonoCamera(Camera):
    """
    This class implements the context manager protocol and simplifies the
    handling of a mono camera.

    It provides the camera tree node by calling :meth:`get_node` or lets you
    directly access an ``NxLibItem`` within the camera node by using the ``[]``
    operator.

    Args:
        serial (str): The serial number of the target camera.
        open_params (dict): Optional parameters for opening the target camera.

    Raises:
        CameraTypeError: If the camera with the given serial number is not a
            monocular camera.
        CameraNotFoundError: If no camera was found for the given serial number.
        CameraOpenError: If the camera with the given serial cannot be opened.
    """
    def __init__(self, serial, open_params={}):
        super().__init__(serial, open_params)
        self._check_type(VAL_MONOCULAR)


class StereoCamera(Camera):
    """
    This class implements the context manager protocol and simplifies the
    handling of an Ensenso stereo camera.

    It provides the camera tree node by calling :meth:`get_node` or lets you
    directly access an ``NxLibItem`` within the camera node by using the ``[]``
    operator.

    Args:
        serial (str): The serial number of the target camera.
        open_params (dict): Optional parameters for opening the target camera.

    Raises:
        CameraTypeError: If the camera with the given serial number is not a
            stereo camera.
        CameraNotFoundError: If no camera was found for the given serial number.
        CameraOpenError: If the camera with the given serial cannot be opened.
    """
    def __init__(self, serial, open_params={}):
        super().__init__(serial, open_params)
        self._check_type(VAL_STEREO)

    def compute_disparity_map(self):
        """
        Compute the disparity map (requires :meth:`capture` to be called first).

        Returns:
            ``NxLibItem``: The disparity map node.
        """
        self._execute(CMD_COMPUTE_DISPARITY_MAP)
        return self._node[ITM_IMAGES][ITM_DISPARITY_MAP]

    def compute_point_map(self):
        """
        Compute the point map (requires :meth:`compute_disparity_map` to be
        called first).

        Returns:
            ``NxLibItem``: The point map node.
        """
        self._execute(CMD_COMPUTE_POINT_MAP)
        return self._node[ITM_IMAGES][ITM_POINT_MAP]

    def compute_texture(self):
        """
        Compute the texture (requires :meth:`compute_point_map` to be called
        first).

        Returns:
            ``NxLibItem``: The texture node.
        """
        self._execute(CMD_COMPUTE_TEXTURE)
        return self._node[ITM_IMAGES][ITM_TEXTURE]

    def get_disparity_map(self):
        """
        The computed disparity map (requires :meth:`compute_disparity_map`).

        Returns:
            `Object`: A byte buffer containing the disparity map.
        """
        return self._node[ITM_IMAGES][ITM_DISPARITY_MAP].get_binary_data()

    def get_point_map(self):
        """
        The computed point map (requires :meth:`compute_point_map`).

        Returns:
            `Object`: A byte buffer containing the point map.
        """
        return self._node[ITM_IMAGES][ITM_POINT_MAP].get_binary_data()

    def get_texture(self):
        """
        The computed texture (requires :meth:`compute_texture`).

        Returns:
            `Object`: A byte buffer containing the texture.
        """
        return self._node[ITM_IMAGES][ITM_TEXTURE].get_binary_data()


class StructuredLightCamera(StereoCamera):
    """
    This class implements the context manager protocol and simplifies the
    handling of an Ensenso structured light camera and has the same
    functionality as a stereo camera except that it does not have a disparity
    map.

    It provides the camera tree node by calling :meth:`get_node` or lets you
    directly access an ``NxLibItem`` within the camera node by using the ``[]``
    operator.

    Args:
        serial (str): The serial number of the target camera.
        open_params (dict): Optional parameters for opening the target camera.

    Raises:
        CameraTypeError: If the camera with the given serial number is not a
            structured light camera.
        CameraNotFoundError: If no camera was found for the given serial number.
        CameraOpenError: If the camera with the given serial cannot be opened.
        CameraDisparityMapError: If a disparity map is requested.
    """
    def __init__(self, serial, open_params={}):
        Camera.__init__(self, serial, open_params)
        self._check_type(VAL_STRUCTURED_LIGHT)

    def compute_disparity_map(self):
        """
        Does nothing, because a structured light camera does not have a
        disparity map. Existing for compatibility reasons.
        """
        pass

    def compute_point_map(self):
        """
        Compute the point map (requires :meth:`capture` to be called first).

        Returns:
            ``NxLibItem``: The point map node.
        """
        super().compute_disparity_map()
        return super().compute_point_map()

    def get_disparity_map(self):
        raise CameraDisparityMapError("A structured light camera does not have"
                                      "a disparity map.")


class CameraTypeError(Exception):
    """ Raised if camera has the wrong type (Mono/Stereo/StructuredLight). """
    pass


class CameraNotFoundError(Exception):
    """ Raised if no camera is found for a given serial number. """
    pass


class CameraOpenError(Exception):
    """ Raised if a camera cannot be opened. """
    pass


class CameraDisparityMapError(Exception):
    """ Raised if a non-existing disparity map is requested. """
    pass



# "from ... import ..." does not work here because of a circular dependency
# between nxlib.api and this module's classes.

import traceback

import nxlib

__all__ = ["NxLibError", "NxLibException"]


class NxLibError(Exception):
    """
    This class represents an NxLib API error.

    Args:
        message (str, optional): The error messgae to be displayed.
            Defaults to None.
    """
    def __init__(self, message=None):
        self._message = message

    def __str__(self):
        return self._message


class NxLibException(NxLibError):
    """
    This class encapsulates NxLib API errors. All methods of
    :class:`~nxlib.item.NxLibItem` and :class:`~nxlib.command.NxLibCommand` not
    taking a return code pointer will throw an NxLibException when the API
    return code indicates an error.

    It is possible to store the ``NxLibCommand`` object that caused the
    exception to keep temporary slots alive while an exception exists.

    Args:
        path (str): The path to the NxLibItem that caused the exception.
        error_code (int): The NxLib error code.
        command (``NxLibCommand``, optional): The command object that cause the
            exception. Defaults to None.
    """
    def __init__(self, path, error_code, command=None):
        self._path = path
        self._error_code = error_code
        # Save command object in the exception to keep temporary slots alive
        # while an exception exists.
        self._command = command

    def get_error_code(self):
        """
        Retrieves the API return code that has caused this exception.

        Returns:
            int: The error code because of which the exception was raised.
        """
        return self._error_code

    def get_error_text(self):
        """
        Retrieves the error text, corresponding to the API return code that has
        caused this exception. This is the error text return by
        :meth:`nxlib.api.translate_error_code`.

        Returns:
            int: The text corresponding to the error code.
        """
        return nxlib.api.translate_error_code(self._error_code)

    def get_item_path(self):
        """
        Retrieves the path of the item that was attempted to access when the
        exception was raised.

        Returns:
            str: The path of the item that has caused the exception.
        """
        return self._path

    def __str__(self):
        message = (f"NxLib error {self._error_code} ({self.get_error_text()}) "
                   f"for item {self._path}")
        try:
            if self._path:
                json_str = nxlib.NxLibItem(self._path).as_json(True)
                message += f"\nCurrent item value: {json_str}"
        except NxLibError:
            pass
        return message


def _handle_exception(exc_type, exc_value, exc_tb):
    if exc_type is None:
        return False
    if exc_type.__name__.startswith("NxLib"):
        print(f"\nAn NxLibError occured: Error Text: {exc_value}\n")
        print("Traceback (most recent call last):")
        traceback.print_tb(exc_tb)
    else:
        print("\nAn non-NxLib related exception occured:\n")
        traceback.print_exception(exc_type, exc_value, exc_tb)
    return True



import os
import pathlib
import platform
import re
import shutil

from nxlib.exception import NxLibError


def get_os():
    os_abbreviations = {"Windows": "win",
                        "Linux": "linux",
                        "Darwin": "mac"}
    os_ = platform.system()
    if not os_:
        return "UndeterminedOperatingSystem"
    elif os_ not in os_abbreviations:
        return os_
    return os_abbreviations[os_]


def _get_architecture_bits():
    bits, _ = platform.architecture()
    return bits


def _get_lib_name(is_remote_lib=False):
    if is_remote_lib:
        default_lib_names = {"win32bit": "NxLibRemote32.dll",
                             "win64bit": "NxLibRemote64.dll",
                             "linux32bit": "libNxLibRemote32.so",
                             "linux64bit": "libNxLibRemote64.so"}
    else:
        default_lib_names = {"win32bit": "NxLib32.dll",
                             "win64bit": "NxLib64.dll",
                             "linux32bit": "libNxLib32.so",
                             "linux64bit": "libNxLib64.so"}
    os_ = get_os()
    bits = _get_architecture_bits()

    try:
        return default_lib_names[os_ + bits]
    except KeyError:
        raise NxLibError(f"Unsupported operating system: {bits}-bit {os_}")


def get_lib_path(is_remote_lib=False):
    path = ""
    if "NXLIB_LIBRARY_PATH" in os.environ:
        path = os.environ["NXLIB_LIBRARY_PATH"] + "/"
    return path + _get_lib_name(is_remote_lib)


def get_default_profiler():
    profiler_path = None
    os_ = get_os()
    if os_ == "win":
        installer_path = pathlib.Path(os.environ["ENSENSO_INSTALL"])
        profiler_path = installer_path / "bin/NxProfiler.exe"
    elif os_ == "linux":
        profiler_path = pathlib.Path("/usr/bin/nxProfiler")
    else:
        raise NxLibError(f"Unsupported operating system: {os_}")
    if not profiler_path.is_file():
        raise NxLibError(f"Profiler executable not found in default path: "
                         f"{profiler_path}")
    return profiler_path


def fix_string_encoding(path):
    try:
        if path is None:
            return b''
        path = path.encode()
    except AttributeError:
        pass
    return path


def convert_camel_to_lower_snake(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def convert_camel_to_upper_snake(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).upper()


def fix_nxlib_prefix(word):
    return word[:2] + word[3:]


def unique_numbered_filepath(path, filename_pattern):
    idx = 0
    while path.joinpath(filename_pattern.format(idx)).exists():
        idx += 1
    return path.joinpath(filename_pattern.format(idx))


def is_filepath(filename):
    """
    Returns True if the given filename is a path to a file and not a sole
    filename. foo return False, foo/bar or foo/bar/file.txt returns True.
    """
    if pathlib.Path(filename).name == filename:
        return False
    else:
        return True


def rmdir(path):
    if not path.is_dir():
        return
    shutil.rmtree(path)



import numpy as np

from nxlib import api
from nxlib.constants import *
from nxlib.exception import NxLibError, NxLibException


__all__ = ["NxLibItem"]


class NxLibItem:
    """
    This class simplifies the concatenation of string and integer constants to
    an NxLib compatible item path specification via its ``[]`` operator.
    Assignment (``=`` and ``<<``) and comparison operators (``<``,  ``<=``,
    ``==``,  ``!=``, ``>`` and ``>=``) simplify the usage of NxLib tree items,
    almost as if using local variables.

    Args:
        path (str, optional): The item's path. If None is given, the item will
            reference the root of the tree. Defaults to None.
    """
    def __init__(self, path=None):
        if path is None:
            path = ""
        self.path = path

    def __getitem__(self, value):
        path = self.path + NXLIB_ITEM_SEPARATOR
        if type(value) is str:
            return NxLibItem(path + value)
        elif type(value) is int:
            return NxLibItem(path + NXLIB_INDEX_ESCAPE_CHAR + str(value))
        else:
            raise NxLibError("Value cannot be added to NxLib path")

    def __setitem__(self, path, value):
        """
        The ``=`` operator. Set the item's value at the given path to the given
        value.

        Args:
            path (str): The path of the item to be set.
            value (int, str, bool or double): The value to be set.
        """
        self[path]._set_t(value)

    def _compare(self, value):
        item_value = self._as_t()
        if type(item_value) == type(value):
            if item_value == value:
                return 0
            elif item_value < value:
                return -1
            else:
                return 1
        else:
            raise NxLibException(self.path, NXLIB_ITEM_TYPE_NOT_COMPATIBLE)

    def __lt__(self, value):
        """
        The ``<`` comparison operator.

        Args:
            value (int, str, bool or double): The value to compare the item's
                value against.

        Returns:
            bool: True if item's value is less than given value, False
            otherwise.
        """
        return self._compare(value) < 0

    def __le__(self, value):
        """
        The ``<=`` comparison operator.

        Args:
            value (int, str, bool or double): The value to compare the item's
                value against.

        Returns:
            bool: True if item's value is less or eqaul to given value,
            False otherwise.
        """
        return self._compare(value) <= 0

    def __eq__(self, value):
        """
        The ``==`` comparison operator.

        Args:
            value (int, str, bool or double): The value to compare the item's
                value against.

        Returns:
            bool: True if item's value is eqaul to given value, False otherwise.
        """
        return self._compare(value) == 0

    def __ne__(self, value):
        """
        The ``!=`` comparison operator.

        Args:
            value (int, str, bool or double): The value to compare the item's
                value against.

        Returns:
            bool: True if item's value is not eqaul to given value, False
            otherwise.
        """
        return self._compare(value) != 0

    def __gt__(self, value):
        """
        The ``>`` comparison operator.

        Args:
            value (int, str, bool or double): The value to compare the item's
                value against.

        Returns:
            bool: True if item's value is greater than given value, False
            otherwise.
        """
        return self._compare(value) > 0

    def __ge__(self, value):
        """
        The ``>=`` comparison operator.

        Args:
            value (int, str, bool or double): The value to compare the item's
                value against.

        Returns:
            bool: True if item's value is greater or eqaul to given value,
            False otherwise.
        """
        return self._compare(value) >= 0

    def __lshift__(self, other):
        """
        The ``<<`` operator.

        Sets a new value for the item, where ``other`` is either a JSON
        formatted string whose value is taken or another NxLibItem whose JSON
        string is taken.

        Args:
            other (str or `NxLibItem`): A JSON formatted string or another
                NxLibItem.

        Raises:
            NxLibException: If ``other`` is neither a string nor an NxLibItem.
        """
        if type(other) is str:
            self.set_json(other, True)
        elif isinstance(other, NxLibItem):
            self.set_json(other.as_json(), True)
        else:
            raise NxLibException(self.path, NXLIB_ITEM_TYPE_NOT_COMPATIBLE)

    def _set_t(self, value):
        if value is None:
            self.set_null()
        elif type(value) is int:
            if (value > 2147483647 or value < -2147483648):
                raise NotImplementedError()
            else:
                self.set_int(value)
        elif type(value) is str:
            self.set_string(value)
        elif type(value) is bool:
            self.set_bool(value)
        elif type(value) is float:
            self.set_double(value)
        else:
            raise NxLibException(self.path, NXLIB_ITEM_TYPE_NOT_COMPATIBLE)

    def set_null(self):
        """ Sets an item to the value Null. """
        api.set_null(self.path)

    def set_double(self, value):
        """
        Sets an item to the given double value.

        Args:
            value (double): Float value to be set.
        """
        api.set_double(self.path, value)

    def set_int(self, value):
        """
        Sets an item to the given int value.

        Args:
            value (int): Integer value to be set.
        """
        api.set_int(self.path, value)

    def set_bool(self, value):
        """
        Sets an item to the given boolean value.

        Args:
            value (bool): Boolean value to be set.
        """
        api.set_bool(self.path, value)

    def set_string(self, value):
        """
        Sets an item to the given string value.

        Args:
            value (string): String value to be set.
        """
        api.set_string(self.path, value)

    def set_json(self, value, only_writable_nodes=False):
        """
        Sets an item to the given JSON value. The value might itself be an
        entire tree structure which will be placed under the specified node.

        Args:
            value (JSON string): The JSON string representing the value or
                subtree to write.
            only_writable_nodes (bool, optional): Specifies whether the function
                should try to write each single node into the existing tree
                structure instead of replacing the entire subtree. When
                specifying True here the function will not complain if certain
                nodes could not be written due to access restrictions. This can
                be used to restore the state of the library or the Parameters
                node of a camera if the tree state has previously been captured
                via :meth:`~as_json`. When this parameter is set to false, all
                items of the entire subtree must not be access restricted in
                order for the function call to succeed! If this parameter is set
                and the given JSON value does not match the current structure of
                the node, the function will return the error code
                ``NxLibItemTypeNotCompatible``. This indicates, that there was
                not a single node with the same structure as in the JSON string
                and the function could never modify any value. In this case you
                probably tried to apply the JSON value to the wrong path.
                Defaults to False.
        """
        api.set_json(self.path, value, only_writable_nodes)

    def set_binary_data(self, buffer, buffer_size_or_width=0, height=0,
                        channel_count=0, bytes_per_element=0, is_float=0):
        """
        Sets data of a ``Binary`` item with either an OpenCV matrix or another
        array object (e.g. Numpy array). In the first case, this function calls
        :meth:`~set_binary_data_from_cv`. In the second case it sets
        the data formatted if ``channel_count`` is non-zero and uses
        ``buffer_size_or_width`` as width, otherwise it sets the data
        unformatted and uses ``buffer_size_or_width`` as buffer size.

        Args:
            buffer (``Object``): Either a OpenCV matrix or another array object
                (e.g. a Numpy array).
            buffer_size_or_width (int, optional): Either the buffer size or the
                array width. Defaults to 0.
            height (int, optional): [description]. The array height to 0.
            channel_count (int): Number of channels for a single item.
                Defaults to 0.
            bytes_per_element (int, optional): Size in bytes of a single channel
                value of an item. Defaults to 0.
            is_float (int, optional): Specifies whether the element data type is
                a floating point type. Defaults to 0.
        """
        if buffer_size_or_width == 0:
            self.set_binary_data_from_cv(buffer)
        elif (channel_count > 0):
            width = buffer_size_or_width
            api.set_binary_formatted(self.path, buffer, width, height,
                                     channel_count, bytes_per_element,
                                     is_float)
        else:
            buffer_size = buffer_size_or_width
            api.set_binary(self.path, buffer, buffer_size)

    def set_binary_data_from_cv(self, mat):
        """
        Sets data of a ``Binary`` item with an OpenCV matrix.

        Args:
            mat (OpenCV matrix): The matrix containing the data to set the item
                content to.

        Raises:
            NxLibException: With NxLib Error ``NxLibItemTypeNotCompatible`` if
                ``mat`` is not an ``ndarray``.
        """
        if type(mat).__name__ != 'ndarray':
            raise NxLibException(self.path, NXLIB_ITEM_TYPE_NOT_COMPATIBLE)

        is_float = False
        if mat.dtype == 'uint8' or mat.dtype == 'int8':
            bytes_per_element = 1
        elif mat.dtype == 'uint16' or mat.dtype == 'int16':
            bytes_per_element = 2
        elif mat.dtype == 'int32':
            bytes_per_element = 4
        elif mat.dtype == 'float32':
            bytes_per_element = 4
            is_float = True
        elif mat.dtype == 'float64':
            bytes_per_element = 8
            is_float = True

        buffer = np.ctypeslib.as_ctypes(mat)
        width, height, channel_count = mat.shape
        api.set_binary_formatted(self.path, buffer, width, height,
                                 channel_count, bytes_per_element, is_float)

    def get_binary_data(self):
        """
        Retrieves data of a ``Binary`` item.

        Returns:
            ``Object``: A byte buffer containing the binary data.
        """
        buffer, buffer_size = self._create_buffer()
        cbuffer = np.ctypeslib.as_ctypes(buffer)
        api.get_binary(self.path, cbuffer, buffer_size)
        return buffer

    def _create_buffer(self):
        binary_data_info = self.get_binary_data_info()
        bytes_per_element, is_float = binary_data_info[3], binary_data_info[4]
        nptype = np.uint8
        if is_float:
            if bytes_per_element == 4:
                nptype = np.float32
            elif bytes_per_element == 8:
                nptype = np.float64
        else:
            if bytes_per_element == 1:
                nptype = np.uint8
            elif bytes_per_element == 2:
                nptype = np.int16
            elif bytes_per_element == 4:
                nptype = np.int32
        width, height, channel_count = binary_data_info[:3]
        buffer = np.zeros((height, width, channel_count), nptype, order='C')
        buffer_size = width * height * channel_count * buffer.dtype.itemsize
        return buffer, buffer_size

    def get_binary_data_info(self):
        """
        Retrieves meta data of a ``Binary`` item.

        Returns:
            tuple containing

            * width (int): The width of the array (consecutive elements in
              memory).
            * height (int): The height of the array (number of rows of ``width``
              * ``channel_count`` elements)
            * channel_count (int): Number of channels for a single item.
            * bytes_per_element (int): Size in bytes of a single channel value
              of an item.
            * is_float (bool): Specifies whether the element data type is a
              floating point type.
            * timestamp (str): The current ``timestamp`` of the binary blob
              queried.
        """
        return api.get_binary_info(self.path)

    def _as_t(self):
        if self.is_null():
            return None
        elif self.is_int():
            return self.as_int()
        elif self.is_double():
            return self.as_double()
        elif self.is_string():
            return self.as_string()
        elif self.is_bool():
            return self.as_bool()
        else:
            raise NxLibException(self.path, NXLIB_ITEM_TYPE_NOT_COMPATIBLE)

    def as_int(self):
        """
        Returns the item value as int.

        Raises:
            NxLibException: With NxLib Error ``NxLibItemTypeNotCompatible`` if
                item is not an int.

        Returns:
            int: The item value.
        """
        return api.get_int(self.path)

    def as_bool(self):
        """
        Returns the item value as bool.

        Raises:
            NxLibException: With NxLib Error ``NxLibItemTypeNotCompatible`` if
                item is not a bool.

        Returns:
            bool: The item value.
        """
        return api.get_bool(self.path)

    def as_double(self):
        """
        Returns the item value as double.

        Raises:
            NxLibException: With NxLib Error ``NxLibItemTypeNotCompatible`` if
                item is not a double.

        Returns:
            double: The item value.
        """
        return api.get_double(self.path)

    def as_string(self):
        """
        Returns the item value as string.

        Raises:
            NxLibException: With NxLib Error ``NxLibItemTypeNotCompatible`` if
                item is not a string.

        Returns:
            str: The item value.
        """
        return api.get_string(self.path)

    def count(self):
        """
        Retrieves the number of subitems of an ``Object`` or the number of
        elements of an ``item``. In case of a ``Binary`` item its data size in
        bytes is returned.

        Returns:
            int: The number of subitems.
        """
        return api.get_count(self.path)

    def as_json(self, pretty_print=1, number_precision=18,
                scientific_number_format=0):
        """
        Retrieves an item value or an entire subtree in JSON representation.

        Args:
            pretty_print (int, optional): Specifies whether to use pretty
                printing. Int is treated as bool. Defaults to 1.
            number_precision (int, optional): The floating point precision of
                the returned numbers. Defaults to 2.
            scientific_number_format (int, optional): Specifies whether to use
                scientific notation for all numbers. Int is treated as bool.
                Defaults to 0.

        Returns:
            JSON string: A JSON string representing the item value or subtree.
        """
        return api.get_json(self.path, pretty_print, number_precision,
                            scientific_number_format)

    def as_json_meta(self, num_levels=1, pretty_print=1, number_precision=18,
                     scientific_number_format=0):
        """
        Retrieves an item value or an entire subtree in JSON representation
        including item metadata (protection, extended type, internal flags).

        Args:
            num_levels (int, optional): The depth of the returned subtree. Nodes
                in lower levels will be omitted. Defaults to 1.
            pretty_print (int, optional): Specifies whether to use pretty
                printing. Int is treated as bool. Defaults to 1.
            number_precision (int, optional): The floating point precision of
                the returned numbers. Defaults to 2.
            scientific_number_format (int, optional): Specifies whether to use
                scientific notation for all numbers. Int is treated as bool.
                Defaults to 0.

        Returns:
            JSON string: A JSON string representing the item value or subtree
            including each item's metadata.
        """
        return api.get_json_meta(self.path, num_levels, pretty_print,
                                 number_precision, scientific_number_format)

    def type(self):
        """
        Retrieves the item type of a tree item.

        Returns:
            int: The type identifier of the basic type of the item.
            See :mod:`~api.constants` for constants named
            ``NXLIB_ITEM_TYPE_*``
        """
        return api.get_type(self.path)

    def is_null(self):
        """
        Returns:
            bool: True if item is null, False otherwise.
        """
        return self.type() == NXLIB_ITEM_TYPE_NULL

    def is_string(self):
        """
        Returns:
            bool: True if item is a string, False otherwise.
        """
        return self.type() == NXLIB_ITEM_TYPE_STRING

    def is_number(self):
        """
        Returns:
            bool: True if item is a number (int or double), False otherwise.
        """
        return self.type() == NXLIB_ITEM_TYPE_NUMBER

    def is_int(self):
        """
        Returns:
            bool: True if item is an integer, False otherwise.
        """
        return self.is_number() and self.as_double().is_integer()

    def is_double(self):
        """
        Returns:
            bool: True if item is a double, False otherwise.
        """
        return self.is_number() and not self.as_double().is_integer()

    def is_bool(self):
        """
        Returns:
            bool: True if item is a bool, False otherwise.
        """
        return self.type() == NXLIB_ITEM_TYPE_BOOL

    def is_array(self):
        """
        Returns:
            bool: True if item is an array, False otherwise.
        """
        return self.type() == NXLIB_ITEM_TYPE_ARRAY

    def is_object(self):
        """
        Returns:
            bool: True if item is an object, False otherwise.
        """
        return self.type() == NXLIB_ITEM_TYPE_OBJECT

    def exists(self):
        """
        Returns:
            bool: True if item exists, False otherwise.
        """
        return api.has_path(self.path)

    def name(self):
        """

        Returns:
            str: The item name.
        """
        return api.get_name(self.path)

    def erase(self):
        """ Erases the item from the tree. """
        api.erase(self.path)

    def wait_for_change(self):
        """
        Wait for any change of the item, i.e. a change of value or item type.
        """
        api.wait_for_change(self.path)

    def wait_for_type(self, awaited_type, wait_for_equal):
        """
        Wait for a type change of the item. When the condition to wait for is
        already satisfied initially, the function returns immediately.

        Args:
            awaited_type (item_type): The item type to wait for.
            wait_for_equal (bool): If True, the function waits until the
                item has the specified type, otherwise it waits until the item
                has a different type than the one specified.
        """
        api.wait_for_type(self.path, awaited_type, wait_for_equal)

    def wait_for_int_value(self, value, wait_for_equal):
        """
        Wait for specific integer value of the item. When the condition to wait
        for is already satisfied initially, the function returns immediately.

        Args:
            value (int): The value to wait for.
            wait_for_equal (bool): If True, the function waits until the item
                has the specified type, otherwise it waits until the item has a
                different type than the one specified.
        """
        api.wait_for_int_value(self.path, value, wait_for_equal)

    def wait_for_string_value(self, value, wait_for_equal):
        """
        Wait for specific string value of the item. When the condition to wait
        for is already satisfied initially, the function returns immediately.

        Args:
            value (str): The value to wait for.
            wait_for_equal (bool): If True, the function waits until the item
                has the specified type, otherwise it waits until the item has a
                different type than the one specified.
        """
        api.wait_for_string_value(self.path, value, wait_for_equal)

    def wait_for_bool_value(self, value, wait_for_equal):
        """
        Wait for specific boolean value of the item. When the condition to wait
        for is already satisfied initially, the function returns immediately.

        Args:
            value (bool): The value to wait for.
            wait_for_equal (bool): If True, the function waits until the item
                has the specified type, otherwise it waits until the item has a
                different type than the one specified.
        """
        api.wait_for_bool_value(self.path, value, wait_for_equal)

    def wait_for_double_value(self, value, wait_for_equal):
        """
        Wait for specific double value the item. When the condition to wait
        for is already satisfied initially, the function returns immediately.

        Args:
            value (double): The value to wait for.
            wait_for_equal (bool): If True, the function waits until the item
                has the specified type, otherwise it waits until the item has a
                different type than the one specified.
        """
        api.wait_for_double_value(self.path, value, wait_for_equal)

    def make_unique_item(self, item_name=None):
        """
        Creates a new item with a unique name.

        Args:
            item_name (str, optional): The parent path of the new item to
                create. Defaults to None.

        Returns:
            str: The name of the generated item.
        """
        new_path = api.make_unique_item(self.path, item_name)
        if new_path:
            return NxLibItem(new_path)
        return NxLibItem()



from pathlib import Path
from subprocess import Popen

import numpy as np

from nxlib import _helper as helper
from nxlib import api
from nxlib.constants import *
from nxlib.item import NxLibItem


NX_LOG_WARNINGS = {
    "NX_LOG_LEVEL_OFF":
    ("Increasing log level from \"Off\" to \"Trace\", otherwise nothing "
     "will be logged by the NxLib."),
    "NX_LOG_OVERFLOW":
    ("A debug buffer overflow occured. Consider incrementing the thread "
     "ring buffer size via [ITM_DEBUG][ITM_BUFFER_SIZE] or a faster buffer "
     "readout."),
    "NX_LOG_IGNORE_NESTED_FILENAME":
    ("Ignoring filename for nested log. Log files are only written for "
     "NxLog objects on first nesting level."),
}


def warn(s):
    if NxLog._print_warnings:
        warning_text = NX_LOG_WARNINGS[s]
        print(f"WARNING: {s}: {warning_text}")


class NxLog:
    """
    This class offers a simple to use interface for logging NxLib events. It
    implements the context manager protocol and thus can be used in a
    ``with`` statement, which encloses the code to be logged. This has the
    nice side effect, that the code to be logged becomes a nested block of code
    that visually contrasts from its surrounding.

    All first-level NxLog objects produce a log file that can be opened with
    NxProfiler. These log files are by default serially numbered and placed in
    a folder named ``nxlog/`` relative to the executing script. The output path
    can be changed for all upcoming NxLog objects with the
    :meth:`~set_output_path` method. Resetting the output path is can be done
    by calling the same method with None.

    By default the NxLib log level is ``NX_LOG_LEVEL_TRACE``. It can either be
    set globally for all NxLog instances with the :meth:`~set_log_level` method
    or individually for each instance with the ``level`` parameter in the
    constructor. Note that the ``NX_LOG_LEVEL_OFF`` will be ignored and
    ``NX_LOG_LEVEL_TRACE`` will be used instead, because otherwise nothing
    will be logged.

    NxLog instances can be nested to further increase the granularity of event
    grouping. Each nested ``with`` statement opens a new level of indentation
    and should be called with a block name, otherwise there won't be any
    grouping within the log file.

    Note:
        All nested NxLog objects log to its corresponding first-level/parent
        object, which finally writes the log file to disk when its scope ends.

    If desired, the written log file is automatically opened with the NxProfiler
    executable if the ``open_in_profiler`` parameter of the constructor was set
    to True or the opening was globally enabled with the
    :meth:`enable_open_in_profiler` class method.

    With the :meth:`~write_debug_message` class method user defined debug
    messages can be written to the log. The message will automatically belong to
    the currently open debug block if there is one.

    Args:
        level (str, optional): The NxLib log level. Defaults to
            ``NX_LOG_LEVEL_TRACE``.
        filename (str, optional): The NxLog filename. If none is given, a
            unique serially numbered filename is created. The ``.nxlog``
            extension is added if missing.  Defaults to None.
        open_in_profiler (bool, optional): If True, the written log file is
            automatically opened with the NxProfiler executable. Defaults to
            False.
    """
    # -------------------------------------------------------------------------
    # Static variables
    # -------------------------------------------------------------------------
    _parent = None
    _log_level = None
    _default_output_path = Path("nxlog/")
    _output_path = _default_output_path
    _open_in_profiler = None
    _print_warnings = True

    # -------------------------------------------------------------------------
    # Constructor
    # -------------------------------------------------------------------------
    def __init__(self, level=None, filename=None, open_in_profiler=False):
        self._parent = self._get_parent()
        self._log_level = self._get_log_level(level)
        self._filename = self._get_filename(filename)
        self._open_in_profiler = self._get_open_in_profiler(open_in_profiler)
        self._content = b''
        self._set_nxlib_log_level()

    # -------------------------------------------------------------------------
    # Implemented protocol methods
    # -------------------------------------------------------------------------
    def __enter__(self):
        # Return self here so that the object assigned by the "as" part of the
        # "with" statement is not None. Necessary for the tests.
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._append_bytes(NxLog.get_debug_buffer())

        # Child exit
        if not self._is_parent():
            return

        # Parent exit
        NxLog._parent = None
        filepath = self._get_filepath()
        filepath.parent.mkdir(parents=True, exist_ok=True)
        self._write_log_file(filepath)
        self._open_profiler(filepath)

    # -------------------------------------------------------------------------
    # Public methods
    # -------------------------------------------------------------------------
    @staticmethod
    def set_log_level(log_level):
        """
        Sets the log level globally for all instances to the given value. Reset
        the log level by either calling this method with ``log_level=None`` or
        ``log_level=NX_LOG_LEVEL_OFF``.

        Args:
            log_level (int): The log level to be set.
        """
        if log_level == NX_LOG_LEVEL_OFF:
            log_level = None
        NxLog._log_level = log_level

    @staticmethod
    def set_output_path(output_path):
        """
        Sets the output path globally for all instances to the given path. Reset
        the output path by calling this method with ``output_path=None``.

        Args:
            output_path (str or pathlib.Path): The output path to be set.
        """
        if output_path is None:
            output_path = NxLog._default_output_path
        NxLog._output_path = Path(output_path)

    @staticmethod
    def enable_open_in_profiler():
        """
        Enables automatic opening of the created log files with NxProfiler.
        """
        NxLog._open_in_profiler = True

    @staticmethod
    def disable_open_in_profiler():
        """
        Disables automatic opening of the created log files with NxProfiler.
        """
        NxLog._open_in_profiler = None

    @staticmethod
    def enable_warnings():
        """ Enables printing of warnings. """
        NxLog._print_warnings = True

    @staticmethod
    def disable_warnings():
        """ Disables printing of warnings. """
        NxLog._print_warnings = False

    @staticmethod
    def write_debug_message(message):
        """
        Inserts a user defined debug message into the NxLib debug stream. The
        message will belong to the last opened debug block if there is any.

        Args:
            message (str): The message to be written.
        """
        api.write_debug_message(message)

    @staticmethod
    def get_debug_buffer():
        """
        This method uses :func:`nxlib.api.get_debug_buffer` to retrieve all
        bytes from the debug buffer. Therefore the API function is called twice,
        once to determine the number of bytes held by the debug buffer and a
        second time to actually read and clear the debug buffer.

        Note:
            A warning is printed if the debug buffer overflowed and warnings
            are enabled. In this case the debug buffer size should be increased
            via the tree item :doc:`/Debug/BufferSize <tree:debug/buffersize>`.

        Returns:
            bytes: The debug buffers content as bytes.
        """
        buffer_size = 0
        for clear_read in [np.int32(0), np.int32(1)]:
            buffer = np.zeros(buffer_size, dtype=np.uint8, order="C")
            cbuffer = np.ctypeslib.as_ctypes(buffer)
            _, buffer_size, overflow = api.get_debug_buffer(
                cbuffer, buffer_size, clear_read)
        if overflow:
            warn("NX_LOG_OVERFLOW")
        return buffer.tobytes()

    # -------------------------------------------------------------------------
    # Private methods
    # -------------------------------------------------------------------------
    def _get_parent(self):
        if NxLog._parent is None:
            NxLog._parent = self
        return NxLog._parent

    def _is_parent(self):
        return self == NxLog._parent

    def _get_log_level(self, level):
        if level:
            if level == NX_LOG_LEVEL_OFF:
                warn("NX_LOG_LEVEL_OFF")
                level = NX_LOG_LEVEL_TRACE
            return level
        elif NxLog._log_level:
            return NxLog._log_level
        else:
            return NX_LOG_LEVEL_TRACE

    def _set_nxlib_log_level(self):
        NxLibItem()[ITM_DEBUG][ITM_LEVEL].set_string(self._log_level)

    def _get_filename(self, filename):
        if filename is None:
            return None
        if filename and not self._is_parent():
            warn("NX_LOG_IGNORE_NESTED_FILENAME")
        # Strip trailing slashes regardless of the OS.
        return str(Path(filename))

    def _get_open_in_profiler(self, open_in_profiler):
        if open_in_profiler or NxLog._open_in_profiler:
            return True
        else:
            return False

    def _append_bytes(self, bytes_):
        if self._is_parent():
            self._content += bytes_
        else:
            self._parent._append_bytes(bytes_)

    def _get_default_filepath(self):
        pattern = "temp-{:04d}.nxlog"
        return helper.unique_numbered_filepath(self._output_path, pattern)

    def _get_filepath(self):
        if self._filename is None:
            return self._get_default_filepath()

        if helper.is_filepath(self._filename):
            return Path(self._filename)

        filepath = NxLog._output_path.joinpath(self._filename)
        if not filepath.suffix == ".nxlog":
            filepath = filepath.with_suffix(".nxlog")
        return filepath

    def _write_log_file(self, filepath):
        with filepath.open(mode="wb") as f:
            f.write(self._content)

    def _open_profiler(self, filepath):
        if self._open_in_profiler:
            cmd = f"{helper.get_default_profiler().resolve()} {str(filepath)}"
            Popen(cmd.split())


class NxDebugBlock(NxLog):
    """
    This class offers a simple to use interface for opening an NxLib debug
    block. It is derived from the :class:`nxlib.log.NxLog` class and thus can be
    used as this to create a log file. It also implements the context manager
    protocol and thus can be used in a ``with`` statement, which encloses the
    code to be logged. This has the nice side effect, that the code to be logged
    becomes a nested block of code that visually contrasts from its surrounding.

    Args:
        name (str, optional): The name of the debug block. Defaults to None.
        level (str, optional): The NxLib log level. Defaults to
            ``NX_LOG_LEVEL_TRACE``.
        filename (str, optional): The NxLog filename. If none is given, a
            unique serially numbered filename is created. The ``.nxlog``
            extension is added if missing.  Defaults to None.
        open_in_profiler (bool, optional): If True, the written log file is
            automatically opened with the NxProfiler executable. Defaults to
            False.
    """
    def __init__(self, name, level=None, filename=None, open_in_profiler=False):
        super().__init__(level, filename, open_in_profiler)
        self._name = name

    def __enter__(self):
        api.open_debug_block(self._name, self._log_level)
        # Return self here so that the object assigned by the "as" part of the
        # "with" statement is not None. Necessary for the tests.
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        api.close_debug_block()
        super().__exit__(exc_type, exc_value, exc_tb)



import sys
from pathlib import Path
from setuptools import setup


def abs_script_dir():
    """ Return the absolute path to the directory this script is located in. """
    return Path(__file__).resolve().parent


def get_version_number():
    """
    Extract version number from VERSION.md if it exists, otherwise extract it
    from the Jenkinsfile in the SDK root folder. The VERSION.md file is created
    by Jenkins and contains ENSENSO_SDK_VERSION.
    """
    version_file = abs_script_dir() / "VERSION.md"
    if version_file.is_file():
        with version_file.open("rt") as f:
            return f.read().strip()

    jenkins_file = abs_script_dir().parent / "Jenkinsfile"
    major, minor = "", ""
    major_str, minor_str = "sdkVersionMajor", "sdkVersionMinor"
    if not jenkins_file.is_file():
        p = jenkins_file.resolve()
        raise FileNotFoundError(f"Jenkinsfile not found in {p}")
    with jenkins_file.open("rt") as f:
        for line in f:
            if major_str in line:
                major = line.split("=")[1].strip()
            if minor_str in line:
                minor = line.split("=")[1].strip()
            if minor and major:
                return major + "." + minor + ".0"
    raise NameError(f"{major_str} or {minor_str} not found in Jenkinsfile")


readme_file = abs_script_dir() / "README.md"
with readme_file.open(encoding='utf-8') as f:
    long_description = f.read()


package_name = "nxlib"
if "--package-name" in sys.argv:
    package_name = sys.argv[2]
    del sys.argv[1:3]


setup(name=package_name,
      packages=["nxlib", "ensenso_nxlib"],
      python_requires=">3.5.0",
      version=get_version_number(),
      description="Python interface to interact with the Ensenso NxLib",
      long_description=long_description,
      long_description_content_type="text/markdown",
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 3",
      ],
      url="https://www.ensenso.com/python-api",
      author="Optonic GmbH",
      author_email="support@optonic.com",
      license="MIT",
      install_requires=[
          "numpy",
      ],
      zip_safe=False,
      extras_require={
      })
